"""Provides methods for handling queries."""

import datetime
import json
import re
from collections.abc import Callable
from typing import Any, TypeVar

from botocore.exceptions import ClientError
from disease.schemas import NAMESPACE_TO_SYSTEM_URI as DISEASE_NAMESPACE_TO_SYSTEM_URI
from disease.schemas import NamespacePrefix as DiseaseNamespacePrefix
from ga4gh.core.models import (
    Coding,
    ConceptMapping,
    Extension,
    MappableConcept,
    Relation,
    code,
)
from uvicorn.config import logger

from therapy import NAMESPACE_LUIS, PREFIX_LOOKUP, SOURCES
from therapy.database import AbstractDatabase
from therapy.schemas import (
    NAMESPACE_TO_SYSTEM_URI,
    SYSTEM_URI_TO_NAMESPACE,
    BaseNormalizationService,
    HasIndication,
    MatchesNormalized,
    MatchType,
    NamespacePrefix,
    NormalizationService,
    RefType,
    SearchService,
    ServiceMeta,
    SourceName,
    SourcePriority,
    Therapy,
    UnmergedNormalizationService,
)

NormService = TypeVar("NormService", bound=BaseNormalizationService)


class InvalidParameterError(Exception):
    """Exception for invalid parameter args provided by the user."""


class QueryHandler:
    """Class for normalizer management. Stores reference to database instance and
    normalizes query input.
    """

    def __init__(self, database: AbstractDatabase) -> None:
        """Initialize QueryHandler instance. Requires a created database object to
        initialize. The most straightforward way to do this is via the ``create_db``
        method in the ``therapy.database`` module:

        >>> from therapy.query import QueryHandler
        >>> from therapy.database import create_db
        >>> q = QueryHandler(create_db())

        :param database: storage backend to search against
        """
        self.db = database

    def _emit_char_warnings(self, query_str: str) -> list[dict]:
        """Emit warnings if query contains non breaking space characters.

        :param str query_str: query string
        :return: List of warnings (dicts)
        """
        warnings: list[dict[str, str]] = []
        nbsp = re.search("\xa0|&nbsp;", query_str)
        if nbsp:
            warnings = [
                {
                    "non_breaking_space_characters": "Query contains non-breaking space characters"
                }
            ]
            logger.warning(
                f"Query ({query_str}) contains non-breaking space characters."
            )
        return warnings

    @staticmethod
    def _get_indication(indication_string: str) -> HasIndication:
        """Load indication data.
        :param str indication_string: dumped JSON string from db
        :return: complete HasIndication object
        """
        indication_values = json.loads(indication_string)
        return HasIndication(
            disease_id=indication_values[0],
            disease_label=indication_values[1],
            normalized_disease_id=indication_values[2],
            supplemental_info=indication_values[3],
        )

    def _add_record(
        self, response: dict[str, dict], item: dict, match_type: str
    ) -> tuple[dict, str]:
        """Add individual record (i.e. Item in DynamoDB) to response object

        :param Dict[str, Dict] response: in-progress response object
        :param Dict item: Item retrieved from DynamoDB
        :param MatchType match_type: type of query match
        :return: Tuple containing updated response object, and string containing name of
            the source of the match
        """
        inds = item.get("has_indication")
        if inds:
            item["has_indication"] = [self._get_indication(i) for i in inds]

        drug = Therapy(**item)
        src_name = item["src_name"]

        matches = response["source_matches"]
        if src_name not in matches:
            pass
        elif matches[src_name] is None:
            matches[src_name] = {
                "match_type": MatchType[match_type.upper()],
                "records": [drug],
                "source_meta_": self.db.get_source_metadata(src_name),
            }
        elif (matches[src_name]["match_type"] == MatchType[match_type.upper()]) and (
            drug.concept_id not in [r.concept_id for r in matches[src_name]["records"]]
        ):
            matches[src_name]["records"].append(drug)

        return response, src_name

    def _fetch_records(
        self, response: dict[str, dict], concept_ids: set[str], match_type: str
    ) -> tuple[dict, set]:
        """Return matched Drug records as a structured response for a given collection
        of concept IDs.

        :param Dict[str, Dict] response: in-progress response object
        :param List[str] concept_ids: List of concept IDs to build from.  Should be all
            lower-case.
        :param str match_type: record should be assigned this type of match.
        :return: response Dict with records filled in via provided concept IDs, and Set
            of source names of matched records
        """
        matched_sources = set()
        for concept_id in concept_ids:
            try:
                match = self.db.get_record_by_id(
                    concept_id.lower(), case_sensitive=False
                )
                if not match:
                    msg = f"Unable to retrieve record for {concept_id}"
                    raise KeyError(msg)
                (response, src) = self._add_record(response, match, match_type)
                matched_sources.add(src)
            except ClientError as e:
                logger.error(e.response["Error"]["Message"])

        return response, matched_sources

    def _fill_no_matches(self, resp: dict[str, Any]) -> dict:
        """Fill all empty source_matches slots with NO_MATCH results.

        :param Dict[str, Dict] resp: incoming response object
        :return: response object with empty source slots filled with NO_MATCH results
            and corresponding source metadata
        """
        for src_name in resp["source_matches"]:
            if resp["source_matches"][src_name] is None:
                resp["source_matches"][src_name] = {
                    "match_type": MatchType.NO_MATCH,
                    "records": [],
                    "source_meta_": self.db.get_source_metadata(src_name),
                }
        return resp

    def _infer_namespace(self, query: str) -> tuple[dict, dict] | None:
        """Retrieve concept ID by inferring namespace. Attempts to match given query
        against known LUI patterns and performs concept ID lookup for all matches.
        :param str query: user-provided query string
        :return: Either tuple containing complete record and warnings if successful,
        or None if unsuccessful
        """
        inferred_records = []
        namespace = None
        for pattern, source in NAMESPACE_LUIS:
            match = re.match(pattern, query)
            if match:
                if source == SourceName.DRUGSATFDA.value:
                    subspace, lui = match.groups()
                    namespace = f"drugsatfda.{subspace.lower()}"
                    inferred_id = f"{namespace}:{lui}"
                else:
                    namespace = NamespacePrefix[source.upper()].value
                    inferred_id = f"{namespace}:{query}"
                record = self.db.get_record_by_id(inferred_id, case_sensitive=False)
                if record:
                    inferred_records.append((record, namespace, inferred_id))
        if inferred_records and namespace:
            inferred_records.sort(key=lambda r: self._record_order(r[0]))
            return (
                inferred_records[0][0],
                {
                    "inferred_namespace": namespace,
                    "adjusted_query": inferred_records[0][2],
                    # probably not possible but just in case
                    "alternate_inferred_matches": [i[2] for i in inferred_records[1:]],
                },
            )
        return None

    def _check_concept_id(
        self, query: str, resp: dict, sources: set[str], infer: bool = True
    ) -> tuple[dict, set]:
        """Check query for concept ID match. Should only find 0 or 1 matches.

        :param str query: search string
        :param Dict resp: in-progress response object to return to client
        :param Set[str] sources: remaining unmatched sources
        :param bool infer: if true, try to infer namespaces for IDs
        :return: Tuple with updated resp object and updated set of unmatched sources
        """
        records = []
        if infer:
            infer_response = self._infer_namespace(query)
            if infer_response:
                records.append(infer_response[0])
                resp["warnings"].append(infer_response[1])
        query_lower = query.lower()
        if [p for p in PREFIX_LOOKUP if query_lower.startswith(p)]:
            record = self.db.get_record_by_id(query, False)
            if record:
                records.append(record)
        for item in records:
            (resp, src_name) = self._add_record(resp, item, MatchType.CONCEPT_ID.name)
            sources = sources - {src_name}
        return resp, sources

    def _check_match_type(
        self, query: str, resp: dict, sources: set[str], match_type: RefType
    ) -> tuple[dict, set]:
        """Check query for selected match type.

        :param query: search string
        :param resp: in-progress response object to return to client
        :param sources: remaining unmatched sources
        :param match_type: Match type to check for
        :return: Tuple with updated resp object and updated set of unmatched sources
        """
        matching_ids = self.db.get_refs_by_type(query, match_type)
        if matching_ids:
            (resp, matched_srcs) = self._fetch_records(
                resp, set(matching_ids), match_type
            )
            sources = sources - matched_srcs
        return resp, sources

    def _get_search_response(
        self, query: str, sources: set[str], infer: bool = True
    ) -> dict:
        """Return response as dict where key is source name and value
        is a list of records.

        :param str query: string to match against
        :param Set[str] sources: sources to match from
        :param bool infer: if true, attempt to infer namespaces from IDs
        :return: completed response object to return to client
        """
        response: dict[str, None | str | list[dict] | dict] = {
            "query": query,
            "warnings": self._emit_char_warnings(query),
            "source_matches": {source: None for source in sources},
        }
        if query == "":
            return self._fill_no_matches(response)
        query = query.strip()

        # check if concept ID match
        response, sources = self._check_concept_id(query, response, sources, infer)
        if len(sources) == 0:
            return response

        query = query.lower()
        for match_type in RefType:
            response, sources = self._check_match_type(
                query, response, sources, match_type
            )
            if len(sources) == 0:
                return response

        # remaining sources get no match
        return self._fill_no_matches(response)

    def search(
        self,
        query_str: str,
        incl: str = "",
        excl: str = "",
        infer: bool = True,
    ) -> SearchService:
        """Fetch normalized therapy objects.

        :param str query_str: query, a string, to search for
        :param str incl: str containing comma-separated names of sources to use. Will
            exclude all other sources. Case-insensitive.
        :param str excl: str containing comma-separated names of source to exclude. Will
            include all other source. Case-insensitive.
        :param bool infer: if true, try to infer namespaces using known Local Unique
            Identifier patterns
        :return: dict containing all matches found in sources.
        :raises InvalidParameterException: if both incl and excl args are provided, or
            if invalid source names are given.
        """
        sources = {}
        sources = {k: v for k, v in SOURCES.items() if self.db.get_source_metadata(v)}
        if not incl and not excl:
            query_sources = set(sources.values())
        elif incl and excl:
            detail = "Cannot request both source inclusions and exclusions."
            raise InvalidParameterError(detail)
        elif incl:
            req_sources = [n.strip() for n in incl.split(",")]
            invalid_sources = []
            query_sources = set()
            for source in req_sources:
                if source.lower() in sources:
                    query_sources.add(sources[source.lower()])
                else:
                    invalid_sources.append(source)
            if invalid_sources:
                detail = f"Invalid source name(s): {invalid_sources}"
                raise InvalidParameterError(detail)
        else:
            req_exclusions = [n.strip() for n in excl.lower().split(",")]
            req_excl_dict = {r.lower(): r for r in req_exclusions}
            invalid_sources = []
            query_sources = set()
            for req_l, req in req_excl_dict.items():
                if req_l not in sources:
                    invalid_sources.append(req)
            for src_l, src in sources.items():
                if src_l not in req_excl_dict:
                    query_sources.add(src)
            if invalid_sources:
                detail = f"Invalid source name(s): {invalid_sources}"
                raise InvalidParameterError(detail)

        response = self._get_search_response(query_str, query_sources, infer)

        response["service_meta_"] = ServiceMeta(
            response_datetime=datetime.datetime.now(tz=datetime.UTC),
        ).model_dump()
        return SearchService(**response)

    def _add_merged_meta(self, response: NormalizationService) -> NormalizationService:
        """Add source metadata to response object.

        :param NormalizationService response: in-progress response object
        :return: completed response object.
        """
        sources_meta = {}
        therapy = response.therapy

        sources = []
        for m in therapy.mappings or []:
            ns = SYSTEM_URI_TO_NAMESPACE.get(m.coding.system)
            if ns in PREFIX_LOOKUP:
                sources.append(PREFIX_LOOKUP[ns])

        for src in sources:
            if src not in sources_meta:
                sources_meta[src] = self.db.get_source_metadata(src)
        response.source_meta_ = sources_meta  # type: ignore[assignment]
        return response

    def _record_order(self, record: dict) -> tuple[int, str]:
        """Construct priority order for matching. Only called by sort().

        :param Dict record: individual record item in iterable to sort
        :return: tuple with rank value and concept ID
        """
        src = record["src_name"].upper()
        source_rank = SourcePriority[src]
        return source_rank, record["concept_id"]

    def _add_therapy(
        self,
        response: NormalizationService,
        record: dict,
        match_type: MatchType,
    ) -> NormalizationService:
        """Format received DB record as Mappable Concept and update response object.
        :param NormalizationService response: in-progress response object
        :param Dict record: record as stored in DB
        :param str query: query string from user request
        :param MatchType match_type: type of match achieved
        :return: completed response object ready to return to user
        """

        def _create_concept_mapping(
            concept_id: str,
            relation: Relation,
            ns_to_system_uri: dict[str, str],
            ns_prefix: NamespacePrefix | DiseaseNamespacePrefix,
        ) -> ConceptMapping:
            """Create concept mapping for therapy or disease identifier

            ``system`` will use OBO Foundry persistent URL (PURL), source homepage, or
            namespace prefix, in that order of preference, if available.

            :param concept_id: Concept identifier represented as a curie
            :param relation: SKOS mapping relationship, default is relatedMatch
            :param ns_to_system_uri: Dictionary containing mapping from namespace to
                system URI
            :param ns_prefix: Namespace prefix enum
            :return: Concept mapping for therapy or disease identifier
            """
            source = concept_id.split(":")[0]

            try:
                source = ns_prefix(source)
            except ValueError:
                try:
                    source = ns_prefix(source.upper())
                except ValueError as e:
                    err_msg = f"Namespace prefix not supported: {source}"
                    raise ValueError(err_msg) from e

            system = ns_to_system_uri.get(source, source)

            return ConceptMapping(
                coding=Coding(code=code(concept_id), system=system), relation=relation
            )

        therapy_obj = MappableConcept(
            id=f"normalize.therapy.{record['concept_id']}",
            primaryCode=code(root=record["concept_id"]),
            conceptType="Therapy",
            label=record.get("label"),
        )

        # mappings
        mappings = [
            _create_concept_mapping(
                concept_id=record["concept_id"],
                relation=Relation.EXACT_MATCH,
                ns_to_system_uri=NAMESPACE_TO_SYSTEM_URI,
                ns_prefix=NamespacePrefix,
            )
        ]
        source_ids = record.get("xrefs", []) + record.get("associated_with", [])
        mappings.extend(
            _create_concept_mapping(
                concept_id=source_id,
                relation=Relation.RELATED_MATCH,
                ns_to_system_uri=NAMESPACE_TO_SYSTEM_URI,
                ns_prefix=NamespacePrefix,
            )
            for source_id in source_ids
        )
        if mappings:
            therapy_obj.mappings = mappings

        extensions = []
        if "aliases" in record:
            extensions.append(Extension(name="aliases", value=record["aliases"]))

        if any(
            filter(
                lambda f: f in record,
                ("approval_ratings", "approval_year", "has_indication"),
            )
        ):
            approv_value = {}
            if "approval_ratings" in record:
                value = record.get("approval_ratings")
                if value:
                    approv_value["approval_ratings"] = value
            if "approval_year" in record:
                value = record.get("approval_year")
                if value:
                    approv_value["approval_year"] = value

            inds = record.get("has_indication", [])
            inds_list = []
            for ind_db in inds:
                indication = self._get_indication(ind_db)

                if indication.normalized_disease_id:
                    mappings = [
                        _create_concept_mapping(
                            concept_id=indication.normalized_disease_id,
                            relation=Relation.RELATED_MATCH,
                            ns_to_system_uri=DISEASE_NAMESPACE_TO_SYSTEM_URI,
                            ns_prefix=DiseaseNamespacePrefix,
                        )
                    ]
                else:
                    mappings = []
                ind_disease_obj = MappableConcept(
                    id=indication.disease_id,
                    conceptType="Disease",
                    label=indication.disease_label,
                    mappings=mappings or None,
                )

                if indication.supplemental_info:
                    ind_disease_obj.extensions = [
                        Extension(name=k, value=v)
                        for k, v in indication.supplemental_info.items()
                    ]
                inds_list.append(ind_disease_obj.model_dump(exclude_none=True))
            if inds_list:
                approv_value["has_indication"] = inds_list

            approv = Extension(name="regulatory_approval", value=approv_value)
            extensions.append(approv)

        trade_names = record.get("trade_names")
        if trade_names:
            extensions.append(Extension(name="trade_names", value=trade_names))

        if extensions:
            therapy_obj.extensions = extensions

        response.match_type = match_type
        response.therapy = therapy_obj
        return self._add_merged_meta(response)

    def _resolve_merge(
        self,
        response: NormService,
        query: str,
        record: dict,
        match_type: MatchType,
        callback: Callable,
    ) -> NormService:
        """Given a record, return the corresponding normalized record

        :param NormalizationService response: in-progress response object
        :param str query: exact query as provided by user
        :param Dict record: record to retrieve normalized concept for
        :param MatchType match_type: type of match that returned these records
        :param Callable callback: response constructor method
        :return: Normalized response object
        """
        merge_ref = record.get("merge_ref")
        if merge_ref:
            # follow merge_ref
            merge = self.db.get_record_by_id(merge_ref, False, True)
            if merge is None:
                logger.error(
                    f"Merge ref lookup failed for ref {record['merge_ref']} "
                    f"in record {record['concept_id']} from query `{query}`"
                )
                return response
            return callback(response, merge, match_type)
        # record is sole member of concept group
        return callback(response, record, match_type)

    def _prepare_normalized_response(self, query: str) -> dict[str, Any]:
        """Provide base response object for normalize endpoints.

        :param str query: user-provided query
        :return: basic normalization response boilerplate
        """
        return {
            "query": query,
            "match_type": MatchType.NO_MATCH,
            "warnings": self._emit_char_warnings(query),
            "service_meta_": ServiceMeta(
                response_datetime=datetime.datetime.now(tz=datetime.UTC)
            ),
        }

    def normalize(self, query: str, infer: bool = True) -> NormalizationService:
        """Return merged, normalized concept for given search term.

        :param str query: string to search against
        :param bool infer: if true, try to infer namespace for IDs
        :return: Normalized response object
        """
        # prepare basic response
        response = NormalizationService(**self._prepare_normalized_response(query))

        return self._perform_normalized_lookup(
            response, query, infer, self._add_therapy
        )

    def _construct_drug_match(self, record: dict) -> Therapy:
        """Create individual Drug match for unmerged normalization endpoint.

        :param Dict record: record to add
        :return: completed Drug object
        """
        inds = record.get("has_indication")
        if inds:
            record["has_indication"] = [self._get_indication(i) for i in inds]
        return Therapy(**record)

    def _add_normalized_records(
        self,
        response: UnmergedNormalizationService,
        normalized_record: dict,
        match_type: MatchType,
    ) -> UnmergedNormalizationService:
        """Add individual records to unmerged normalize response.

        :param UnmergedNormalizationService response: in-progress response object
        :param Dict normalized_record: record associated with normalized concept,
            either merged or single identity
        :param MatchType match_type: type of match achieved
        :return: Completed response object
        """
        response.match_type = match_type
        response.normalized_concept_id = normalized_record["concept_id"]
        if normalized_record["item_type"] == "identity":
            record_source = SourceName[normalized_record["src_name"].upper()]
            response.source_matches[record_source] = MatchesNormalized(
                records=[self._construct_drug_match(normalized_record)],
                source_meta_=self.db.get_source_metadata(record_source),
            )
        else:
            concept_ids = [
                normalized_record["concept_id"],
                *normalized_record.get("xrefs", []),
            ]
            for concept_id in concept_ids:
                record = self.db.get_record_by_id(concept_id, case_sensitive=False)
                if not record:
                    continue  # cover a few chemidplus edge cases
                record_source = SourceName[record["src_name"].upper()]
                drug = self._construct_drug_match(record)
                if record_source in response.source_matches:
                    response.source_matches[record_source].records.append(drug)
                else:
                    response.source_matches[record_source] = MatchesNormalized(
                        records=[drug],
                        source_meta_=self.db.get_source_metadata(record_source),
                    )
        return response

    def _perform_normalized_lookup(
        self, response: NormService, query: str, infer: bool, response_builder: Callable
    ) -> NormService:
        """Retrieve normalized concept, for use in normalization endpoints
        :param NormService response: in-progress response object
        :param str query: user-provided query
        :param bool infer: whether to try namespace inference
        :param Callable response_builder: response constructor callback method
        :return: completed service response object
        """
        if query == "":
            return response
        query_str = query.lower().strip()

        # check merged concept ID match
        record = self.db.get_record_by_id(query_str, case_sensitive=False, merge=True)
        if record:
            return response_builder(response, record, MatchType.CONCEPT_ID)

        # check concept ID match
        record = self.db.get_record_by_id(query_str, case_sensitive=False)
        if record:
            return self._resolve_merge(
                response, query, record, MatchType.CONCEPT_ID, response_builder
            )

        # check concept ID match with inferred namespace
        if infer:
            inferred_response = self._infer_namespace(query)
            if inferred_response:
                if response.warnings:
                    response.warnings.append(inferred_response[1])
                else:
                    response.warnings = [inferred_response[1]]
                return self._resolve_merge(
                    response,
                    query,
                    inferred_response[0],
                    MatchType.CONCEPT_ID,
                    response_builder,
                )

        # check other match types
        for match_type in RefType:
            matching_refs = self.db.get_refs_by_type(query_str, match_type)
            matching_records = [
                self.db.get_record_by_id(ref, False) for ref in matching_refs
            ]
            matching_records.sort(key=self._record_order)  # type: ignore[arg-type]

            # attempt merge ref resolution until successful
            for match in matching_records:
                if match is None:
                    raise ValueError
                record = self.db.get_record_by_id(match["concept_id"], False)
                if record:
                    match_type_value = MatchType[match_type.upper()]
                    return self._resolve_merge(
                        response, query, record, match_type_value, response_builder
                    )

        return response

    def normalize_unmerged(
        self, query: str, infer: bool = True
    ) -> UnmergedNormalizationService:
        """Return all source records under the normalized concept for the provided
        query string.

        :param str query: string to search against
        :param bool infer: if true, try to infer namespace for IDs
        :return: Normalized response object
        """
        response = UnmergedNormalizationService(
            source_matches={}, **self._prepare_normalized_response(query)
        )
        return self._perform_normalized_lookup(
            response, query, infer, self._add_normalized_records
        )



"""Create concept groups and merged records."""

import logging
import re
from timeit import default_timer as timer
from typing import Any

from tqdm import tqdm

from therapy.database.database import AbstractDatabase, DatabaseWriteError
from therapy.schemas import RefType, SourceName, SourcePriority

logger = logging.getLogger(__name__)


class Merge:
    """Handles record merging."""

    def __init__(self, database: AbstractDatabase, silent: bool = True) -> None:
        """Initialize Merge instance.

        * self._groups is a dictionary keying concept IDs to the Set of concept IDs in
        that group. This is redundantly captured for all group members (i.e., a group
        of 5 concepts would all have their own key, and the value would be an identical
        Set of the same IDs)
        * self._unii_to_drugsatfda keys UNII codes to valid Drugs@FDA concepts.
        Because UNIIs aren't stored directly in groups, we utilize a separate lookup
        table to prevent repeat queries.
        * self._failed_lookups stores concept IDs for which lookups have been attempted
        and failed. Because we don't associate these IDs with groups, a separate
        mapping is necessary to prevent repeat queries.

        :param database: db instance to use for record retrieval and creation.
        :param silent: if True, suppress all console output
        """
        self.database = database
        self._groups: dict[str, set[str]] = {}
        self._unii_to_drugsatfda: dict[str, set[str]] = {}
        self._failed_lookups: set[str] = set()
        self._silent = silent

    def create_merged_concepts(self, record_ids: set[str]) -> None:
        """Create concept groups, generate merged concept records, and update database.

        :param Set[str] record_ids: concept identifiers from which groups should be
            generated.
        """
        logger.info("Generating record ID sets...")
        start = timer()
        for record_id in tqdm(record_ids, ncols=80, disable=self._silent):
            new_group = self._create_record_id_set(record_id)
            if new_group:
                for concept_id in new_group:
                    self._groups[concept_id] = new_group
        end = timer()
        logger.debug("Built record ID sets in %s seconds", end - start)

        self._groups = {k: v for k, v in self._groups.items() if len(v) > 1}

        logger.info("Creating merged records and updating database...")
        uploaded_ids = set()
        start = timer()
        for record_id, group in tqdm(
            self._groups.items(), ncols=80, disable=self._silent
        ):
            if record_id in uploaded_ids:
                continue
            merged_record = self._generate_merged_record(group)

            # add group merger item to DB
            self.database.add_merged_record(merged_record)

            # add updated references
            for concept_id in group:
                merge_ref = merged_record["concept_id"]
                try:
                    self.database.update_merge_ref(concept_id, merge_ref)
                except DatabaseWriteError as dw:
                    if str(dw).startswith("No such record exists"):
                        logger.error(
                            "Updating nonexistent record: %s for merge ref to %s",
                            concept_id,
                            merge_ref,
                        )
                    else:
                        logger.error(str(dw))
            uploaded_ids |= group
        self.database.complete_write_transaction()
        logger.info("Merged concept generation successful.")
        end = timer()
        logger.debug("Generated and added concepts in %s seconds", end - start)

    def _get_drugsatfda_from_unii(self, ref: str) -> str | None:
        """Given an `associated_with` item keying a UNII code to a Drugs@FDA record,
        verify that the record can be safely added to a concept group.
        Drugs@FDA tracks a number of "compound therapies", and provides UNIIs to each
        individual component. If we included them in normalized record sets, they would
        end up merging distinct therapies under the umbrella of the compound group.
        We're excluding Drugs@FDA records with multiple UNIIs as a tentative solution.

        :param ref: a Drugs@FDA concept ID that includes an xref to a UNII identifier
        :return: Drugs@FDA concept ID if record meets rules, None otherwise
        """
        fetched = self.database.get_record_by_id(ref, False)
        if fetched:
            uniis = [
                a for a in fetched.get("associated_with", []) if a.startswith("unii")
            ]
        else:
            logger.error("Couldn't retrieve record for %s", ref)
            return None
        if len(uniis) == 1:
            return fetched["concept_id"]
        return None

    def _get_xrefs(self, record: dict[str, Any]) -> set[str]:
        """Extract references to entries in other sources from a record.

        :param Dict record: record to process
        :return: Set of xref values
        """
        xrefs = set(record.get("xrefs", []))
        unii_xrefs = [
            a for a in record.get("associated_with", []) if a.startswith("unii:")
        ]
        for unii in unii_xrefs:
            # get Drugs@FDA records that are associated_with this UNII
            drugsatfda_ids = self._unii_to_drugsatfda.get(unii)
            if drugsatfda_ids is not None:
                xrefs |= drugsatfda_ids
                continue

            unii_assoc = self.database.get_refs_by_type(
                unii.lower(), RefType.ASSOCIATED_WITH
            )
            drugsatfda_refs = set()
            for ref in unii_assoc:
                if ref.startswith("drugsatfda"):
                    drugsatfda_ref = self._get_drugsatfda_from_unii(ref)
                    if drugsatfda_ref:
                        drugsatfda_refs.add(drugsatfda_ref)
            self._unii_to_drugsatfda[unii] = drugsatfda_refs
            xrefs |= drugsatfda_refs
        return xrefs

    def _create_record_id_set(
        self, record_id: str, observed_id_set: set | None = None
    ) -> set[str]:
        """Create concept ID group for an individual record ID.

        :param str record_id: concept ID for record to build group from
        :return: set of related identifiers pertaining to a common concept.
        """
        if observed_id_set is None:
            observed_id_set = set()

        if record_id in self._groups:
            return self._groups[record_id]
        if record_id.startswith("drugsatfda"):
            return {record_id}
        if record_id in self._failed_lookups:
            return observed_id_set - {record_id}

        # get record
        db_record = self.database.get_record_by_id(record_id)
        if not db_record:
            if record_id.startswith("rxcui"):
                brand_lookup = self.database.get_rxnorm_id_by_brand(record_id)
                if brand_lookup:
                    return self._create_record_id_set(brand_lookup, observed_id_set)
            logger.warning(
                "Unable to resolve lookup for %s in ID set: %s",
                record_id,
                observed_id_set,
            )
            self._failed_lookups.add(record_id)
            return observed_id_set - {record_id}

        # construct group
        local_id_set = self._get_xrefs(db_record)
        if not local_id_set:
            return observed_id_set | {db_record["concept_id"]}
        merged_id_set = {record_id} | observed_id_set
        for local_record_id in local_id_set - observed_id_set:
            merged_id_set |= self._create_record_id_set(local_record_id, merged_id_set)
        return merged_id_set

    def _create_record_id_sets(self, record_ids: set[str]) -> None:
        """Update self._groups with normalized concept groups.
        :param Set[str] record_ids: concept identifiers from which groups should be
            generated.
        """
        logger.info("Generating record ID sets...")
        start = timer()
        for record_id in record_ids:
            new_group = self._create_record_id_set(record_id)
            if new_group:
                for concept_id in new_group:
                    self._groups[concept_id] = new_group
        end = timer()
        logger.debug("Built record ID sets in %s seconds", end - start)

    _biologic_suffix_pattern = re.compile(r"^(.*)[ -][a-z]{4}$")

    def _sort_records(self, records: list[dict]) -> list[dict]:
        """Ensure proper sorting of records in group.

        First, order by source priority and tiebreak by smallest concept ID value.
        Then, if the first entry appears to be an RxNorm biosimilar, find the base
        therapeutic concept and move it in the front. This is necessary to ensure that
        the normalized drug's label is something like "trastuzumab" and not
        "trastuzumab-abcd". See
        https://www.fda.gov/files/drugs/published/Nonproprietary-Naming-of-Biological-Products-Guidance-for-Industry.pdf,
        and https://github.com/cancervariants/therapy-normalization/issues/299.

        This method is broken out to faciliate more direct testing.

        :param records: List of records in normalized group
        :return: sorted records list
        """

        def _record_order(record: dict) -> tuple[int, str]:
            """Provide priority values of concepts for sort function.

            :param record: individual therapy record
            :return: tuple (sortable) of source priority, and then concept ID
            :raise ValueError: if unrecognized source
            """
            src = record["src_name"].upper()
            if src == "DRUGS@FDA":
                src = "DRUGSATFDA"
            if src in SourcePriority.__members__:
                source_rank = SourcePriority[src].value
            else:
                msg = f"Prohibited source: {src} in concept_id {record['concept_id']}"
                raise ValueError(msg)
            return source_rank, record["concept_id"]

        records.sort(key=_record_order)

        if len([r for r in records if r["src_name"] == SourceName.RXNORM]) <= 1:
            return records
        first_match = re.findall(
            self._biologic_suffix_pattern, records[0].get("label", "")
        )
        if first_match:
            base = first_match[0].lower()
            for i, record in enumerate(records[1:], start=1):
                if (
                    record["src_name"] == SourceName.RXNORM
                    and record.get("label", "").lower() == base
                ):
                    logger.debug(
                        "Reordering RxNorm entry %s ahead of biosimilars",
                        record["concept_id"],
                    )
                    main_record = records.pop(i)
                    records = [main_record, *records]
                    break
        return records

    def _generate_merged_record(self, record_id_set: set[str]) -> dict:
        """Generate merged record from provided concept ID group.
        Where attributes are 'set-like', they should be combined, and where
        they are 'scalar-like', assign from the highest-priority source where
        that attribute is not None.

        Uses the SourcePriority schema to define source priority.

        :param Set record_id_set: group of concept IDs
        :return: completed merged drug object to be stored in DB
        """
        records = []
        for record_id in record_id_set:
            record = self.database.get_record_by_id(record_id)
            if record:
                records.append(record)
            else:
                logger.error(
                    "Merge record generator could not retrieve record for %s in %s",
                    record_id,
                    record_id_set,
                )
        records = self._sort_records(records)

        # initialize merged record
        merged_attrs = {
            "concept_id": records[0]["concept_id"],
            "xrefs": [r["concept_id"] for r in records[1:]],
            "label": None,
            "aliases": set(),
            "trade_names": set(),
            "associated_with": set(),
            "approval_ratings": set(),
            "approval_year": set(),
            "has_indication": [],
        }

        # merge from constituent records
        set_fields = ["aliases", "trade_names", "associated_with", "approval_year"]
        for record in records:
            for field in set_fields:
                merged_attrs[field] |= set(record.get(field, set()))
            approval_ratings = record.get("approval_ratings")
            if approval_ratings:
                merged_attrs["approval_ratings"] |= set(approval_ratings)
            label = record.get("label")
            if label:
                if merged_attrs["label"] is None:
                    merged_attrs["label"] = label
                elif label != merged_attrs["label"]:
                    merged_attrs["aliases"].add(label)
            for ind in record.get("has_indication", []):
                if ind not in merged_attrs["has_indication"]:
                    merged_attrs["has_indication"].append(ind)

        # clear unused fields
        for field in [*set_fields, "has_indication", "approval_ratings"]:
            field_value = merged_attrs[field]
            if field_value:
                merged_attrs[field] = list(field_value)
            else:
                del merged_attrs[field]
        if not merged_attrs["label"]:
            del merged_attrs["label"]

        merged_attrs["label_and_type"] = f"{merged_attrs['concept_id'].lower()}##merger"
        merged_attrs["item_type"] = "merger"
        return merged_attrs



"""Build test data for GuideToPharmacology"""

import csv
from pathlib import Path

from therapy.database import create_db
from therapy.etl import GuideToPHARMACOLOGY

TEST_IDS = {"5343", "2169", "2804", "240", "3303", "5260"}

gtop = GuideToPHARMACOLOGY(create_db())
gtop._extract_data(False)
TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "guidetopharmacology"
ligands_file_path = TEST_DATA_DIR / gtop._data_files.ligands.name
mapping_file_path = TEST_DATA_DIR / gtop._data_files.ligand_id_mapping.name

ligands_rows = []
with gtop._data_files.ligands.open() as f:
    reader = csv.reader(f, delimiter="\t")
    ligands_rows.append(next(reader))
    ligands_rows.append(next(reader))

    for row in reader:
        if row[0] in TEST_IDS:
            ligands_rows.append(row)

with ligands_file_path.open("w") as f:
    writer = csv.writer(f, delimiter="\t", quoting=csv.QUOTE_ALL)
    writer.writerows(ligands_rows)

map_rows = []
with gtop._data_files.ligand_id_mapping.open() as f:
    reader = csv.reader(f, delimiter="\t")
    map_rows.append(next(reader))
    map_rows.append(next(reader))

    for row in reader:
        if row[0] in TEST_IDS:
            map_rows.append(row)

with mapping_file_path.open("w") as f:
    writer = csv.writer(f, delimiter="\t", quoting=csv.QUOTE_ALL)
    writer.writerows(map_rows)



"""Test DynamoDB.

In the future, this should be filled in with different tests for alternate
implementations.
"""

from boto3.dynamodb.conditions import Key


def test_tables_created(database):
    """Check that therapy_concepts and therapy_metadata are created."""
    existing_tables = database.dynamodb_client.list_tables()["TableNames"]
    assert "therapy_normalizer" in existing_tables


def test_item_type(database):
    """Check that objects are tagged with item_type attribute."""
    filter_exp = Key("label_and_type").eq("chembl:chembl11359##identity")
    item = database.therapies.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "identity"

    filter_exp = Key("label_and_type").eq("interferon alfacon-1##label")
    item = database.therapies.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "label"

    filter_exp = Key("label_and_type").eq("unii:5x5hb3vz3z##associated_with")
    item = database.therapies.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "associated_with"

    filter_exp = Key("label_and_type").eq("drugbank:db00515##xref")
    item = database.therapies.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "xref"

    filter_exp = Key("label_and_type").eq("spiramycin i##alias")
    item = database.therapies.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "alias"

    filter_exp = Key("label_and_type").eq("rxcui:1041527##rx_brand")
    item = database.therapies.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "rx_brand"

    filter_exp = Key("label_and_type").eq("align##trade_name")
    item = database.therapies.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "trade_name"

    filter_exp = Key("label_and_type").eq("rxcui:9991##merger")
    item = database.therapies.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "merger"



"""A base class for extraction, transformation, and loading of data."""

import contextlib
import json
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import ClassVar

import click
from disease.database import create_db as create_disease_db
from disease.query import QueryHandler as DiseaseNormalizer
from pydantic import ValidationError
from wags_tails import (
    ChemblData,
    ChemIDplusData,
    CustomData,
    DataSource,
    DrugBankData,
    DrugsAtFdaData,
    GToPLigandData,
    HemOncData,
    NcitData,
    RxNormData,
)

from therapy import ITEM_TYPES
from therapy.database import AbstractDatabase
from therapy.etl.rules import Rules
from therapy.schemas import SourceName, Therapy

_logger = logging.getLogger(__name__)


DATA_DISPATCH = {
    SourceName.CHEMBL: ChemblData,
    SourceName.CHEMIDPLUS: ChemIDplusData,
    SourceName.DRUGBANK: DrugBankData,
    SourceName.DRUGSATFDA: DrugsAtFdaData,
    SourceName.GUIDETOPHARMACOLOGY: GToPLigandData,
    SourceName.HEMONC: HemOncData,
    SourceName.NCIT: NcitData,
    SourceName.RXNORM: RxNormData,
}


class EtlError(Exception):
    """Raise for data transform errors."""


class Base(ABC):
    """The ETL base class.

    Default methods are declared to provide basic functions for core source
    data-gathering phases (extraction, transformation, loading).

    Classes should expand or reimplement these methods as needed.
    """

    def __init__(
        self,
        database: AbstractDatabase,
        data_path: Path | None = None,
        silent: bool = True,
    ) -> None:
        """Extract from sources.

        :param database: application database object
        :param data_path: path to app data directory
        :param silent: if True, don't print ETL results to console
        """
        # self._name = SourceName[self.__class__.__name__.upper()]
        self._silent = silent
        self._name = SourceName(self.__class__.__name__)
        self._data_source: (
            ChemblData
            | ChemIDplusData
            | DrugBankData
            | DrugsAtFdaData
            | GToPLigandData
            | HemOncData
            | NcitData
            | RxNormData
            | CustomData
        ) = self._get_data_handler(data_path)  # type: ignore
        self.database = database
        self._added_ids: list[str] = []
        self._rules = Rules(self._name)

    def _get_data_handler(self, data_path: Path | None = None) -> DataSource:
        """Construct data handler instance for source. Overwrite for edge-case sources.

        :param data_path: location of data storage
        :return: instance of wags_tails.DataSource to manage source file(s)
        """
        return DATA_DISPATCH[self._name](data_dir=data_path, silent=self._silent)

    def perform_etl(self, use_existing: bool = False) -> list[str]:
        """Public-facing method to begin ETL procedures on given data.
        Returned concept IDs can be passed to Merge method for computing
        merged concepts.

        :param use_existing: if True, don't try to retrieve latest source data
        :param silent: if True, suppress all console output
        :return: list of concept IDs which were successfully processed and
            uploaded.
        """
        self._extract_data(use_existing)
        if not self._silent:
            click.echo("Transforming and loading data to DB...")
        self._load_meta()
        self._transform_data()
        self.database.complete_write_transaction()
        return self._added_ids

    def _extract_data(self, use_existing: bool) -> None:
        """Acquire source data.

        :param bool use_existing: if True, don't try to fetch latest source data
        """
        self._data_file, self._version = self._data_source.get_latest(
            from_local=use_existing
        )

    @abstractmethod
    def _load_meta(self) -> None:
        """Load source metadata entry."""
        raise NotImplementedError

    @abstractmethod
    def _transform_data(self) -> None:
        """Prepare source data for loading into DB. Individually extract each
        record and call the Base class's `_load_therapy()` method.
        """
        raise NotImplementedError

    @staticmethod
    def _process_searchable_attributes(therapy: dict) -> dict:
        """Apply standardization to searchable therapy fields, e.g. ``label``,
        ``trade_names`` etc

        * delete redundant values
        * sort
        * trim if field has > 20 values
        * remove empty fields

        :param therapy: in-progress therapy object
        :return: processed therapy object
        """
        for attr_type in ITEM_TYPES:
            if attr_type in therapy:
                value = therapy[attr_type]
                if value is None or value == []:
                    del therapy[attr_type]
                    continue

                if attr_type == "label":
                    therapy["label"] = value.strip()
                    continue

                unique_values = {v.strip() for v in value}
                unique_values = set(filter(None, unique_values))
                if attr_type == "aliases" and "trade_names" in therapy:
                    value = list(unique_values - set(therapy["trade_names"]))
                else:
                    value = list(unique_values)

                if (attr_type in ("aliases", "trade_names")) and ("label" in therapy):
                    with contextlib.suppress(ValueError):
                        value.remove(therapy["label"])

                if len(value) > 20:
                    _logger.debug("%s has > 20 %s.", therapy["concept_id"], attr_type)
                    del therapy[attr_type]
                    continue

                value.sort()
                therapy[attr_type] = value
        return therapy

    @staticmethod
    def _indication_sorter(indication: dict) -> str:
        """Produce sortable value for indication object.

        :param indication: ``has_indication`` object
        :return: either given max_phase, or an empty string
        """
        max_phase = indication.get("supplemental_info", {}).get(
            "chembl_max_phase_for_ind"
        )
        # sometimes this value is explicitly set to None, which is unsortable
        if max_phase is None:
            max_phase = ""
        return max_phase

    def _process_detail_fields(self, therapy: dict) -> dict:
        """Apply standardization to therapy detail fields, e.g. ``has_indication``,
        ``approval_year``, ``approval_ratings``.

        ``has_indication`` is sorted first by highest to lowest CHEMBL approval rating,
        if available, and then by given disease ID/label

        :param therapy: in-progress therapy object
        :return: therapy object with detail fields processed appropriately
        """
        indications = therapy.get("has_indication")
        if indications:
            indications.sort(key=self._indication_sorter, reverse=True)
            indications.sort(
                key=lambda x: (x.get("disease_id"), x.get("disease_label"))
            )
            indications = list(
                {
                    json.dumps(
                        [
                            ind["disease_id"],
                            ind["disease_label"],
                            ind.get("normalized_disease_id"),
                            ind.get("supplemental_info"),
                        ]
                    )
                    for ind in indications
                }
            )
            therapy["has_indication"] = indications
        elif "has_indication" in therapy:
            del therapy["has_indication"]

        # handle detail fields
        approval_attrs = ("approval_ratings", "approval_year")
        for field in approval_attrs:
            if approval_attrs in therapy and therapy[field] is None:
                del therapy[field]
        return therapy

    def _load_therapy(self, therapy: dict) -> None:
        """Load individual therapy record into database. This method takes
        responsibility for:
            * validating record structure correctness
            * removing duplicates from and sorting list-like fields
            * removing empty fields

        :param therapy: valid therapy object.
        """
        try:
            Therapy(**therapy)
        except ValidationError as e:
            _logger.error("Attempted to load invalid therapy: %s", therapy)
            raise e

        therapy = self._rules.apply_rules_to_therapy(therapy)
        therapy = self._process_searchable_attributes(therapy)
        therapy = self._process_detail_fields(therapy)

        self.database.add_record(therapy, self._name)
        self._added_ids.append(therapy["concept_id"])


class DiseaseIndicationBase(Base):
    """Base class for sources that require disease normalization capabilities."""

    _disease_cache: ClassVar[dict[str, str | None]] = {}

    def __init__(
        self,
        database: AbstractDatabase,
        data_path: Path | None = None,
        silent: bool = True,
    ) -> None:
        """Initialize source ETL instance.

        :param database: application database object
        :param data_path: path to app data directory
        :param silent: if True, don't print ETL results to console
        """
        super().__init__(database, data_path, silent)
        self.disease_normalizer = DiseaseNormalizer(create_disease_db())

    def _normalize_disease(self, query: str) -> str | None:
        """Attempt normalization of disease term.

        :param str query: term to normalize
        :return: ID if successful, None otherwise
        """
        term = query.lower()
        if term in self._disease_cache:
            return self._disease_cache[term]
        response = self.disease_normalizer.normalize(term)
        normalized_id = response.disease.primaryCode.root if response.disease else None
        self._disease_cache[term] = normalized_id
        if normalized_id is None:
            _logger.warning("Failed to normalize disease term: %s", query)
        return normalized_id


class SourceFormatError(Exception):
    """Raise when source data formatting is incompatible with the source transformation
    methods: for example, if columns in a CSV file have changed.
    """



"""Test merged record generation."""

import json
import os
import random
from pathlib import Path

import pytest

from therapy.database import AWS_ENV_VAR_NAME
from therapy.etl.chembl import ChEMBL
from therapy.etl.chemidplus import ChemIDplus
from therapy.etl.drugbank import DrugBank
from therapy.etl.drugsatfda import DrugsAtFDA
from therapy.etl.guidetopharmacology import GuideToPHARMACOLOGY
from therapy.etl.hemonc import HemOnc
from therapy.etl.merge import Merge
from therapy.etl.ncit import NCIt
from therapy.etl.rxnorm import RxNorm
from therapy.etl.wikidata import Wikidata
from therapy.schemas import SourceName


@pytest.fixture(scope="module")
def merge_instance(test_source, is_test_env, database):
    """Provide fixture for ETL merge class"""
    if is_test_env:
        if os.environ.get(AWS_ENV_VAR_NAME):
            pytest.fail(
                f"Running the full therapy ETL pipeline test on an AWS environment is forbidden -- either unset {AWS_ENV_VAR_NAME} or unset THERAPY_TEST"
            )
        for SourceClass in (  # noqa: N806
            ChEMBL,
            ChemIDplus,
            DrugBank,
            DrugsAtFDA,
            GuideToPHARMACOLOGY,
            HemOnc,
            NCIt,
            RxNorm,
            Wikidata,
        ):
            test_source(SourceClass)
    return Merge(database)


def compare_merged_records(actual: dict, fixture: dict):
    """Check that records are identical."""
    assert actual["concept_id"] == fixture["concept_id"]
    assert actual["label_and_type"] == fixture["label_and_type"]
    assert ("label" in actual) == ("label" in fixture)
    if "label" in actual or "label" in fixture:
        assert actual["label"] == fixture["label"]
    assert ("trade_names" in actual) == ("trade_names" in fixture)
    if "trade_names" in actual or "trade_names" in fixture:
        assert set(actual["trade_names"]) == set(fixture["trade_names"])
    assert ("aliases" in actual) == ("aliases" in fixture)
    if "aliases" in actual or "aliases" in fixture:
        assert set(actual["aliases"]) == set(fixture["aliases"])
    assert ("xrefs" in actual) == ("xrefs" in fixture)
    if "xrefs" in actual or "xrefs" in fixture:
        assert set(actual["xrefs"]) == set(fixture["xrefs"])
    assert ("associated_with" in actual) == ("associated_with" in fixture)
    if "associated_with" in actual or "associated_with" in fixture:
        assert set(actual["associated_with"]) == set(fixture["associated_with"])
    assert ("approval_ratings" in actual) == ("approval_ratings" in fixture)
    if "approval_ratings" in actual or "approval_ratings" in fixture:
        assert set(actual["approval_ratings"]) == set(fixture["approval_ratings"])
    assert ("approval_year" in actual) == ("approval_year" in fixture)
    if "approval_year" in actual or "approval_year" in fixture:
        assert set(actual["approval_year"]) == set(fixture["approval_year"])
    assert ("has_indication" in actual) == ("has_indication" in fixture)
    if "has_indication" in actual or "has_indication" in fixture:
        actual_inds = actual["has_indication"].copy()
        fixture_inds = fixture["has_indication"].copy()
        assert len(actual_inds) == len(fixture_inds)
        actual_inds.sort()
        fixture_inds.sort()
        for i in range(len(actual_inds)):
            assert actual_inds[i] == fixture_inds[i]


@pytest.fixture(scope="module")
def fixture_data(test_data: Path):
    """Fetch fixture data"""
    with (test_data / "fixtures" / "merged_fixtures.json").open() as f:
        return json.load(f)


@pytest.fixture(scope="module")
def phenobarbital_merged(fixture_data) -> dict:
    """Create phenobarbital fixture."""
    return fixture_data["phenobarbital"]


@pytest.fixture(scope="module")
def cisplatin_merged(fixture_data) -> dict:
    """Create cisplatin fixture."""
    return fixture_data["cisplatin"]


@pytest.fixture(scope="module")
def spiramycin_merged(fixture_data) -> dict:
    """Create fixture for spiramycin. The RxNorm entry should be inaccessible
    to this group.
    """
    return fixture_data["spiramycin"]


@pytest.fixture(scope="module")
def amifostine_merged(fixture_data) -> dict:
    """Create fixture for amifostine."""
    return fixture_data["amifostine"]


@pytest.fixture(scope="module")
def record_id_groups():
    """Create fixture for testing concept group creation."""
    groups = [
        # phenobarbital
        {
            "rxcui:8134",
            "ncit:C739",
            "chemidplus:50-06-6",
            "wikidata:Q407241",
            "drugbank:DB01174",
            "chembl:CHEMBL40",
            "iuphar.ligand:2804",
        },
        # spiramycin
        {
            "rxcui:9991",
            "ncit:C839",
            "chemidplus:8025-81-8",
            "drugbank:DB06145",
            "wikidata:Q422265",
        },
        # cisplatin
        {
            "rxcui:2555",
            "ncit:C376",
            "chemidplus:15663-27-1",
            "wikidata:Q412415",
            "drugbank:DB00515",
            "drugbank:DB12117",
            "hemonc:105",
            "chembl:CHEMBL11359",
            "drugsatfda.anda:074656",
            "drugsatfda.anda:074735",
            "drugsatfda.anda:206774",
            "drugsatfda.anda:207323",
            "drugsatfda.anda:075036",
            "iuphar.ligand:5343",
            "drugsatfda.nda:018057",
        },
        # Amifostine
        # tests lookup of wikidata reference to rxnorm brand record, and
        # drugbank reference to dead chemidplus record
        {
            "rxcui:1545987",
            "rxcui:4126",
            "ncit:C488",
            "ncit:C66724",
            "drugbank:DB01143",
            "drugsatfda.nda:020221",
            "chembl:CHEMBL1006",
            "chemidplus:112901-68-5",
            "chemidplus:20537-88-6",
            "wikidata:Q251698",
        },
        # Therapeutic Procedure
        {"ncit:C49236"},
        # test exclusion of drugs@fda records with multiple UNIIs
        {"drugsatfda.nda:210595"},
    ]
    groups_keyed = {}
    for group in groups:
        for concept_id in group:
            groups_keyed[concept_id] = group
    return groups_keyed


def test_id_sets(merge_instance: Merge, record_id_groups: dict[str, set[str]]):
    """Test creation of record ID sets. Queries DB and matches against
    record_id_groups fixture.
    """
    # try a few different orders
    keys = list(record_id_groups)
    key_len = len(keys)
    order0 = list(range(key_len))
    random.seed(42)
    orders = [random.sample(order0, key_len) for _ in range(5)]
    for order in [order0, *orders]:
        ordered_keys = [keys[i] for i in order]
        merge_instance._create_record_id_sets(ordered_keys)  # type: ignore
        groups = merge_instance._groups

        # perform checks
        for concept_id in groups:
            assert groups[concept_id] == record_id_groups[concept_id], f"{concept_id}"
        assert len(groups) == len(record_id_groups)  # check if any are missing

    # test dead reference
    has_dead_ref = "ncit:C107245"
    dead_group = merge_instance._create_record_id_set(has_dead_ref)
    assert dead_group == {has_dead_ref}


def test_generate_merged_record(
    merge_instance: Merge,
    record_id_groups: dict[str, set[str]],
    phenobarbital_merged: dict,
    cisplatin_merged: dict,
    spiramycin_merged: dict,
):
    """Test generation of merged record method."""
    phenobarbital_ids = record_id_groups["rxcui:8134"]
    merge_response = merge_instance._generate_merged_record(phenobarbital_ids)
    compare_merged_records(merge_response, phenobarbital_merged)

    cisplatin_ids = record_id_groups["rxcui:2555"]
    merge_response = merge_instance._generate_merged_record(cisplatin_ids)
    compare_merged_records(merge_response, cisplatin_merged)

    spiramycin_ids = record_id_groups["ncit:C839"]
    merge_response = merge_instance._generate_merged_record(spiramycin_ids)
    compare_merged_records(merge_response, spiramycin_merged)


def test_create_merged_concepts(
    merge_instance: Merge,
    record_id_groups: dict[str, set[str]],
    phenobarbital_merged: dict,
    cisplatin_merged: dict,
    spiramycin_merged: dict,
    amifostine_merged: dict,
    mocker,
):
    """Test end-to-end creation and upload of merged concepts."""
    add_spy = mocker.spy(merge_instance.database, "add_merged_record")
    update_spy = mocker.spy(merge_instance.database, "update_merge_ref")
    merge_instance.create_merged_concepts(set(record_id_groups))
    merge_instance.database.complete_write_transaction()

    # check merged record generation and storage
    added_records = {
        call[1][0]["concept_id"]: call[1][0] for call in add_spy.mock_calls
    }

    phenobarbital_merged_id = phenobarbital_merged["concept_id"]
    assert phenobarbital_merged_id in added_records
    compare_merged_records(added_records[phenobarbital_merged_id], phenobarbital_merged)

    cisplatin_merged_id = cisplatin_merged["concept_id"]
    assert cisplatin_merged_id in added_records
    compare_merged_records(added_records[cisplatin_merged_id], cisplatin_merged)

    spiramycin_merged_id = spiramycin_merged["concept_id"]
    assert spiramycin_merged_id in added_records
    compare_merged_records(added_records[spiramycin_merged_id], spiramycin_merged)

    amifostine_merged_id = amifostine_merged["concept_id"]
    assert amifostine_merged_id in added_records
    compare_merged_records(added_records[amifostine_merged_id], amifostine_merged)

    # should only create new records for groups with n > 1 members
    assert add_spy.call_count == 4

    # check merged record reference updating
    updated_records = {k[1][0]: k[1][1] for k in update_spy.mock_calls}
    for concept_id in record_id_groups["rxcui:8134"]:
        assert updated_records[concept_id] == phenobarbital_merged["concept_id"].lower()
    for concept_id in record_id_groups["rxcui:2555"]:
        assert updated_records[concept_id] == cisplatin_merged["concept_id"].lower()
    for concept_id in record_id_groups["ncit:C839"]:
        assert updated_records[concept_id] == spiramycin_merged["concept_id"].lower()

    # no merged record should be generated
    assert "ncit:C49236" not in updated_records
    assert "drugsatfda.nda:210595" not in updated_records

    assert update_spy.call_count == len(record_id_groups) - 2


def test_merge_record_ordering(merge_instance: Merge):
    """Test record ordering within merged record generation.

    This test uses dummy data and only checks the ordering pipeline because relevant
    edge cases are quite large and would entail considerable effort to maintain fixtures.
    """
    records = [
        {
            "label": "Trastuzumab",
            "src_name": SourceName.HEMONC,
            "concept_id": "hemonc:1",
        },
        {
            "label": "trastuzumab-xxxx",
            "src_name": SourceName.RXNORM,
            "concept_id": "rxcui:1",
        },
        {
            "label": "Trastuzumab",
            "src_name": SourceName.RXNORM,
            "concept_id": "rxcui:2",
        },
    ]
    result = merge_instance._sort_records(records)
    assert [r["concept_id"] for r in result] == ["rxcui:2", "rxcui:1", "hemonc:1"]

    records = [
        {
            "label": "drug",
            "src_name": SourceName.DRUGSATFDA,
            "concept_id": "drugsatfda:1",
        },
        {"label": "drug", "src_name": SourceName.NCIT, "concept_id": "ncit:1"},
        {
            "label": "Trastuzumab-xxxx",
            "src_name": SourceName.RXNORM,
            "concept_id": "rxcui:1",
        },
        {
            "label": "Trastuzumab-yyyy",
            "src_name": SourceName.RXNORM,
            "concept_id": "rxcui:2",
        },
    ]
    result = merge_instance._sort_records(records)
    assert [r["concept_id"] for r in result] == [
        "rxcui:1",
        "rxcui:2",
        "ncit:1",
        "drugsatfda:1",
    ]

    records = [
        {
            "label": "trastuzumab-aaaa",
            "src_name": SourceName.RXNORM,
            "concept_id": "rxcui:2",
        },
        {
            "label": "trastuzumab",
            "src_name": SourceName.CHEMBL,
            "concept_id": "chembl:CHEMBL1",
        },
    ]
    result = merge_instance._sort_records(records)
    assert [r["concept_id"] for r in result] == ["rxcui:2", "chembl:CHEMBL1"]



"""Test the emit_warnings function."""

from therapy.query import QueryHandler


def test_emit_warnings(database):
    """Test that emit_warnings works correctly."""
    expected_warnings = [
        {
            "non_breaking_space_characters": "Query contains non-breaking space characters"
        }
    ]
    query_handler = QueryHandler(database)

    # Test emit no warnings
    actual_warnings = query_handler._emit_char_warnings("CISPLATIN")
    assert actual_warnings == []

    # Test emit warnings
    actual_warnings = query_handler._emit_char_warnings("CIS PLATIN")
    assert actual_warnings == actual_warnings

    actual_warnings = query_handler._emit_char_warnings("CIS\u00a0platin")
    assert expected_warnings == actual_warnings

    actual_warnings = query_handler._emit_char_warnings("CIS&nbsp;platin")
    assert expected_warnings == actual_warnings

    actual_warnings = query_handler._emit_char_warnings("CIS\xa0platin")
    assert expected_warnings == actual_warnings



"""Construct test data for Wikidata"""

import datetime
import json
from pathlib import Path

from wikibaseintegrator.wbi_helpers import execute_sparql_query

from therapy.etl.wikidata import SPARQL_QUERY

TEST_IDS = {
    "http://www.wikidata.org/entity/Q412415",
    "http://www.wikidata.org/entity/Q407241",
    "http://www.wikidata.org/entity/Q26272",
    "http://www.wikidata.org/entity/Q15353101",
    "http://www.wikidata.org/entity/Q191924",
    "http://www.wikidata.org/entity/Q418702",
    "http://www.wikidata.org/entity/Q251698",
    "http://www.wikidata.org/entity/Q27287118",
    "http://www.wikidata.org/entity/Q422265",
}

result = execute_sparql_query(SPARQL_QUERY)["results"]["bindings"]
test_data = []
for item in result:
    if item["item"]["value"] in TEST_IDS:
        params = {}
        for attr in item:
            params[attr] = item[attr]["value"]
        test_data.append(params)

TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "wikidata"
date = datetime.datetime.now(tz=datetime.UTC).strftime("%Y-%m-%d")
outfile_path = TEST_DATA_DIR / f"wikidata_{date}.json"
with outfile_path.open("w+") as f:
    json.dump(test_data, f, indent=2)



"""Build DrugBank test data."""

import csv
from pathlib import Path

from therapy.database import create_db
from therapy.etl import DrugBank

TEST_IDS = [
    "DB06145",
    "DB01143",
    "DB01174",
    "DB12117",
    "DB00515",
    "DB00522",
    "DB14257",
]

db = DrugBank(create_db())
db._extract_data(False)
TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "drugbank"
outfile_path = TEST_DATA_DIR / db._data_file.name

with db._data_file.open() as f:
    rows = list(csv.DictReader(f))

write_rows = []
for row in rows:
    if row["DrugBank ID"] in TEST_IDS:
        write_rows.append(row)

with outfile_path.open("w") as f:
    writer = csv.DictWriter(f, write_rows[0].keys())
    writer.writeheader()
    writer.writerows(write_rows)
